// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    array::{Array, ArrayType},
    bundle::{Bundle, BundleType},
    enum_::{Enum, EnumType, EnumVariant},
    expr::{
        ops::{self, EnumLiteral},
        CastBitsTo, CastToBits, Expr, ExprEnum, HdlPartialEq, ToExpr,
    },
    hdl,
    int::{DynSize, Size, UInt, UIntType},
    intern::{Intern, Interned},
    memory::{DynPortType, Mem, MemPort},
    module::{
        transform::visit::{Fold, Folder},
        Block, Module, NameIdGen, ScopedNameId, Stmt, StmtConnect, StmtIf, StmtMatch, StmtWire,
    },
    source_location::SourceLocation,
    ty::{CanonicalType, Type},
    wire::Wire,
};
use core::fmt;
use hashbrown::HashMap;

#[derive(Debug)]
pub enum SimplifyEnumsError {
    EnumIsNotCastableFromBits { enum_type: Enum },
}

impl fmt::Display for SimplifyEnumsError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            SimplifyEnumsError::EnumIsNotCastableFromBits { enum_type } => write!(
                f,
                "simplify_enums failed: enum type is not castable from bits: {enum_type:?}"
            ),
        }
    }
}

impl std::error::Error for SimplifyEnumsError {}

#[hdl]
struct TagAndBody<T, BodyWidth: Size> {
    tag: T,
    body: UIntType<BodyWidth>,
}

#[derive(Clone, Debug)]
enum EnumTypeState {
    TagEnumAndBody(TagAndBody<Enum, DynSize>),
    TagUIntAndBody(TagAndBody<UInt, DynSize>),
    UInt(UInt),
    Unchanged,
}

struct State {
    enum_types: HashMap<Enum, EnumTypeState>,
    replacement_mem_ports: HashMap<MemPort<DynPortType>, Wire<CanonicalType>>,
    kind: SimplifyEnumsKind,
    name_id_gen: NameIdGen,
}

impl State {
    fn get_or_make_enum_type_state(
        &mut self,
        enum_type: Enum,
    ) -> Result<EnumTypeState, SimplifyEnumsError> {
        if let Some(retval) = self.enum_types.get(&enum_type) {
            return Ok(retval.clone());
        }
        if !enum_type.type_properties().is_castable_from_bits {
            return Err(SimplifyEnumsError::EnumIsNotCastableFromBits { enum_type });
        }
        let has_body = enum_type
            .variants()
            .iter()
            .any(|variant| variant.ty.is_some());
        let retval = match (self.kind, has_body) {
            (SimplifyEnumsKind::SimplifyToEnumsWithNoBody, true) => {
                EnumTypeState::TagEnumAndBody(TagAndBody {
                    tag: Enum::new(Interned::from_iter(enum_type.variants().iter().map(|v| {
                        EnumVariant {
                            name: v.name,
                            ty: None,
                        }
                    }))),
                    body: UInt::new_dyn(
                        enum_type.type_properties().bit_width - enum_type.discriminant_bit_width(),
                    ),
                })
            }
            (SimplifyEnumsKind::SimplifyToEnumsWithNoBody, false) => EnumTypeState::Unchanged,
            (SimplifyEnumsKind::ReplaceWithBundleOfUInts, _) => {
                EnumTypeState::TagUIntAndBody(TagAndBody {
                    tag: UInt::new_dyn(enum_type.discriminant_bit_width()),
                    body: UInt::new_dyn(
                        enum_type.type_properties().bit_width - enum_type.discriminant_bit_width(),
                    ),
                })
            }
            (SimplifyEnumsKind::ReplaceWithUInt, _) => {
                EnumTypeState::UInt(UInt::new_dyn(enum_type.type_properties().bit_width))
            }
        };
        self.enum_types.insert(enum_type, retval.clone());
        Ok(retval)
    }
}

fn connect_port(
    stmts: &mut Vec<Stmt>,
    lhs: Expr<CanonicalType>,
    rhs: Expr<CanonicalType>,
    source_location: SourceLocation,
) {
    if Expr::ty(lhs) == Expr::ty(rhs) {
        stmts.push(
            dbg!(StmtConnect {
                lhs,
                rhs,
                source_location,
            })
            .into(),
        );
        return;
    }
    match (Expr::ty(lhs), Expr::ty(rhs)) {
        (CanonicalType::Bundle(lhs_type), CanonicalType::UInt(_) | CanonicalType::Bool(_)) => {
            let lhs = Expr::<Bundle>::from_canonical(lhs);
            for field in lhs_type.fields() {
                assert!(!field.flipped);
                connect_port(stmts, Expr::field(lhs, &field.name), rhs, source_location);
            }
        }
        (CanonicalType::UInt(_) | CanonicalType::Bool(_), CanonicalType::Bundle(rhs_type)) => {
            let rhs = Expr::<Bundle>::from_canonical(rhs);
            for field in rhs_type.fields() {
                assert!(!field.flipped);
                connect_port(stmts, lhs, Expr::field(rhs, &field.name), source_location);
            }
        }
        (CanonicalType::Bundle(lhs_type), CanonicalType::Bundle(_)) => {
            let lhs = Expr::<Bundle>::from_canonical(lhs);
            let rhs = Expr::<Bundle>::from_canonical(rhs);
            for field in lhs_type.fields() {
                let (lhs_field, rhs_field) = if field.flipped {
                    (Expr::field(rhs, &field.name), Expr::field(lhs, &field.name))
                } else {
                    (Expr::field(lhs, &field.name), Expr::field(rhs, &field.name))
                };
                connect_port(stmts, lhs_field, rhs_field, source_location);
            }
        }
        (CanonicalType::Array(lhs_type), CanonicalType::Array(_)) => {
            let lhs = Expr::<Array>::from_canonical(lhs);
            let rhs = Expr::<Array>::from_canonical(rhs);
            for index in 0..lhs_type.len() {
                connect_port(stmts, lhs[index], rhs[index], source_location);
            }
        }
        (CanonicalType::Bundle(_), _)
        | (CanonicalType::Enum(_), _)
        | (CanonicalType::Array(_), _)
        | (CanonicalType::UInt(_), _)
        | (CanonicalType::SInt(_), _)
        | (CanonicalType::Bool(_), _)
        | (CanonicalType::Clock(_), _)
        | (CanonicalType::AsyncReset(_), _)
        | (CanonicalType::SyncReset(_), _)
        | (CanonicalType::Reset(_), _) => unreachable!(
            "trying to connect memory ports:\n{:?}\n{:?}",
            Expr::ty(lhs),
            Expr::ty(rhs),
        ),
    }
}

impl Folder for State {
    type Error = SimplifyEnumsError;

    fn fold_enum(&mut self, _v: Enum) -> Result<Enum, Self::Error> {
        unreachable!()
    }

    fn fold_module<T: BundleType>(&mut self, v: Module<T>) -> Result<Module<T>, Self::Error> {
        let old_name_id_gen =
            std::mem::replace(&mut self.name_id_gen, NameIdGen::for_module(v.canonical()));
        let retval = Fold::default_fold(v, self);
        self.name_id_gen = old_name_id_gen;
        retval
    }

    fn fold_expr_enum(&mut self, op: ExprEnum) -> Result<ExprEnum, Self::Error> {
        match op {
            ExprEnum::EnumLiteral(op) => Ok(match self.get_or_make_enum_type_state(op.ty())? {
                EnumTypeState::TagEnumAndBody(TagAndBody { tag, body }) => *Expr::expr_enum(
                    <TagAndBody<Enum, DynSize> as BundleType>::Builder::default()
                        .field_tag(EnumLiteral::new_by_index(tag, op.variant_index(), None))
                        .field_body(match op.variant_value() {
                            Some(variant_value) => variant_value.fold(self)?.cast_to_bits(),
                            None => body.zero().to_expr(),
                        })
                        .to_expr(),
                ),
                EnumTypeState::TagUIntAndBody(TagAndBody { tag, body }) => *Expr::expr_enum(
                    <TagAndBody<UInt, DynSize> as BundleType>::Builder::default()
                        .field_tag(tag.from_int_wrapping(op.variant_index()))
                        .field_body(match op.variant_value() {
                            Some(variant_value) => variant_value.fold(self)?.cast_to_bits(),
                            None => body.zero().to_expr(),
                        })
                        .to_expr(),
                ),
                EnumTypeState::UInt(_) => *Expr::expr_enum(
                    <TagAndBody<UInt, DynSize> as BundleType>::Builder::default()
                        .field_tag(
                            UIntType::new(op.ty().discriminant_bit_width())
                                .from_int_wrapping(op.variant_index()),
                        )
                        .field_body(match op.variant_value() {
                            Some(variant_value) => variant_value.fold(self)?.cast_to_bits(),
                            None => UIntType::new(
                                op.ty().type_properties().bit_width
                                    - op.ty().discriminant_bit_width(),
                            )
                            .zero()
                            .to_expr(),
                        })
                        .cast_to_bits(),
                ),
                EnumTypeState::Unchanged => ExprEnum::EnumLiteral(ops::EnumLiteral::new_by_index(
                    op.ty(),
                    op.variant_index(),
                    op.variant_value().map(|v| v.fold(self)).transpose()?,
                )),
            }),
            ExprEnum::VariantAccess(op) => Ok(
                match self.get_or_make_enum_type_state(Expr::ty(op.base()))? {
                    EnumTypeState::TagEnumAndBody(_) | EnumTypeState::TagUIntAndBody(_) => {
                        match op.variant_type() {
                            Some(variant_type) => *Expr::expr_enum(
                                Expr::<TagAndBody<CanonicalType, DynSize>>::from_canonical(
                                    (*Expr::expr_enum(op.base())).fold(self)?.to_expr(),
                                )
                                .body[..variant_type.bit_width()]
                                    .cast_bits_to(variant_type),
                            ),
                            None => *Expr::expr_enum(().to_expr()),
                        }
                    }
                    EnumTypeState::UInt(_) => match op.variant_type() {
                        Some(variant_type) => {
                            let base_int = Expr::<UInt>::from_canonical(
                                (*Expr::expr_enum(op.base())).fold(self)?.to_expr(),
                            );
                            dbg!(base_int);
                            let base_ty = Expr::ty(op.base());
                            let variant_type_bit_width = variant_type.bit_width();
                            *Expr::expr_enum(
                                base_int[base_ty.discriminant_bit_width()..]
                                    [..variant_type_bit_width]
                                    .cast_bits_to(variant_type),
                            )
                        }
                        None => *Expr::expr_enum(().to_expr()),
                    },
                    EnumTypeState::Unchanged => match op.variant_type() {
                        Some(_) => ExprEnum::VariantAccess(ops::VariantAccess::new_by_index(
                            op.base().fold(self)?,
                            op.variant_index(),
                        )),
                        None => *Expr::expr_enum(().to_expr()),
                    },
                },
            ),
            ExprEnum::MemPort(mem_port) => Ok(
                if let Some(&wire) = self.replacement_mem_ports.get(&mem_port) {
                    ExprEnum::Wire(wire)
                } else {
                    ExprEnum::MemPort(mem_port.fold(self)?)
                },
            ),
            ExprEnum::UIntLiteral(_)
            | ExprEnum::SIntLiteral(_)
            | ExprEnum::BoolLiteral(_)
            | ExprEnum::BundleLiteral(_)
            | ExprEnum::ArrayLiteral(_)
            | ExprEnum::Uninit(_)
            | ExprEnum::NotU(_)
            | ExprEnum::NotS(_)
            | ExprEnum::NotB(_)
            | ExprEnum::Neg(_)
            | ExprEnum::BitAndU(_)
            | ExprEnum::BitAndS(_)
            | ExprEnum::BitAndB(_)
            | ExprEnum::BitOrU(_)
            | ExprEnum::BitOrS(_)
            | ExprEnum::BitOrB(_)
            | ExprEnum::BitXorU(_)
            | ExprEnum::BitXorS(_)
            | ExprEnum::BitXorB(_)
            | ExprEnum::AddU(_)
            | ExprEnum::AddS(_)
            | ExprEnum::SubU(_)
            | ExprEnum::SubS(_)
            | ExprEnum::MulU(_)
            | ExprEnum::MulS(_)
            | ExprEnum::DivU(_)
            | ExprEnum::DivS(_)
            | ExprEnum::RemU(_)
            | ExprEnum::RemS(_)
            | ExprEnum::DynShlU(_)
            | ExprEnum::DynShlS(_)
            | ExprEnum::DynShrU(_)
            | ExprEnum::DynShrS(_)
            | ExprEnum::FixedShlU(_)
            | ExprEnum::FixedShlS(_)
            | ExprEnum::FixedShrU(_)
            | ExprEnum::FixedShrS(_)
            | ExprEnum::CmpLtB(_)
            | ExprEnum::CmpLeB(_)
            | ExprEnum::CmpGtB(_)
            | ExprEnum::CmpGeB(_)
            | ExprEnum::CmpEqB(_)
            | ExprEnum::CmpNeB(_)
            | ExprEnum::CmpLtU(_)
            | ExprEnum::CmpLeU(_)
            | ExprEnum::CmpGtU(_)
            | ExprEnum::CmpGeU(_)
            | ExprEnum::CmpEqU(_)
            | ExprEnum::CmpNeU(_)
            | ExprEnum::CmpLtS(_)
            | ExprEnum::CmpLeS(_)
            | ExprEnum::CmpGtS(_)
            | ExprEnum::CmpGeS(_)
            | ExprEnum::CmpEqS(_)
            | ExprEnum::CmpNeS(_)
            | ExprEnum::CastUIntToUInt(_)
            | ExprEnum::CastUIntToSInt(_)
            | ExprEnum::CastSIntToUInt(_)
            | ExprEnum::CastSIntToSInt(_)
            | ExprEnum::CastBoolToUInt(_)
            | ExprEnum::CastBoolToSInt(_)
            | ExprEnum::CastUIntToBool(_)
            | ExprEnum::CastSIntToBool(_)
            | ExprEnum::CastBoolToSyncReset(_)
            | ExprEnum::CastUIntToSyncReset(_)
            | ExprEnum::CastSIntToSyncReset(_)
            | ExprEnum::CastBoolToAsyncReset(_)
            | ExprEnum::CastUIntToAsyncReset(_)
            | ExprEnum::CastSIntToAsyncReset(_)
            | ExprEnum::CastSyncResetToBool(_)
            | ExprEnum::CastSyncResetToUInt(_)
            | ExprEnum::CastSyncResetToSInt(_)
            | ExprEnum::CastSyncResetToReset(_)
            | ExprEnum::CastAsyncResetToBool(_)
            | ExprEnum::CastAsyncResetToUInt(_)
            | ExprEnum::CastAsyncResetToSInt(_)
            | ExprEnum::CastAsyncResetToReset(_)
            | ExprEnum::CastResetToBool(_)
            | ExprEnum::CastResetToUInt(_)
            | ExprEnum::CastResetToSInt(_)
            | ExprEnum::CastBoolToClock(_)
            | ExprEnum::CastUIntToClock(_)
            | ExprEnum::CastSIntToClock(_)
            | ExprEnum::CastClockToBool(_)
            | ExprEnum::CastClockToUInt(_)
            | ExprEnum::CastClockToSInt(_)
            | ExprEnum::FieldAccess(_)
            | ExprEnum::ArrayIndex(_)
            | ExprEnum::DynArrayIndex(_)
            | ExprEnum::ReduceBitAndU(_)
            | ExprEnum::ReduceBitAndS(_)
            | ExprEnum::ReduceBitOrU(_)
            | ExprEnum::ReduceBitOrS(_)
            | ExprEnum::ReduceBitXorU(_)
            | ExprEnum::ReduceBitXorS(_)
            | ExprEnum::SliceUInt(_)
            | ExprEnum::SliceSInt(_)
            | ExprEnum::CastToBits(_)
            | ExprEnum::CastBitsTo(_)
            | ExprEnum::ModuleIO(_)
            | ExprEnum::Instance(_)
            | ExprEnum::Wire(_)
            | ExprEnum::Reg(_) => op.default_fold(self),
        }
    }

    fn fold_block(&mut self, block: Block) -> Result<Block, Self::Error> {
        let mut memories = vec![];
        let mut stmts = vec![];
        for memory in block.memories {
            let old_element_ty = memory.array_type().element();
            let new_element_ty = memory.array_type().element().fold(self)?;
            if new_element_ty != old_element_ty {
                let mut new_ports = vec![];
                for port in memory.ports() {
                    let new_port = MemPort::<DynPortType>::new_unchecked(
                        port.mem_name(),
                        port.source_location(),
                        port.port_name(),
                        port.addr_type(),
                        new_element_ty,
                    );
                    new_ports.push(new_port);
                    let new_port_ty = new_port.ty();
                    let mut wire_ty_fields = Vec::from_iter(new_port_ty.fields());
                    if let Some(wmask_name) = new_port.port_kind().wmask_name() {
                        let index = *new_port_ty
                            .name_indexes()
                            .get(&wmask_name.intern())
                            .unwrap();
                        wire_ty_fields[index].ty = port.ty().fields()[index].ty;
                    }
                    let wire_ty = Bundle::new(Intern::intern_owned(wire_ty_fields));
                    if wire_ty == new_port_ty {
                        continue;
                    }
                    let wire_name = self.name_id_gen.gen(
                        (*format!("{}_{}", memory.scoped_name().1 .0, port.port_name())).intern(),
                    );
                    let wire = Wire::new_unchecked(
                        ScopedNameId(memory.scoped_name().0, wire_name),
                        port.source_location(),
                        wire_ty,
                    );
                    stmts.push(
                        StmtWire {
                            annotations: Default::default(),
                            wire: wire.canonical(),
                        }
                        .into(),
                    );
                    connect_port(
                        &mut stmts,
                        Expr::canonical(new_port.to_expr()),
                        Expr::canonical(wire.to_expr()),
                        port.source_location(),
                    );
                    self.replacement_mem_ports.insert(port, wire.canonical());
                }
                memories.push(Mem::new_unchecked(
                    memory.scoped_name(),
                    memory.source_location(),
                    ArrayType::new_dyn(new_element_ty, memory.array_type().len()),
                    memory.initial_value(),
                    Intern::intern_owned(new_ports),
                    memory.read_latency(),
                    memory.write_latency(),
                    memory.read_under_write(),
                    memory.port_annotations(),
                    memory.mem_annotations(),
                ));
            } else {
                memories.push(memory.fold(self)?);
            }
        }
        stmts.extend_from_slice(&block.stmts.fold(self)?);
        Ok(Block {
            memories: Intern::intern_owned(memories),
            stmts: Intern::intern_owned(stmts),
        })
    }

    fn fold_stmt(&mut self, stmt: Stmt) -> Result<Stmt, Self::Error> {
        fn match_int_tag(
            state: &mut State,
            int_tag_expr: Expr<UInt>,
            source_location: SourceLocation,
            blocks: Interned<[Block]>,
        ) -> Result<StmtIf, SimplifyEnumsError> {
            let mut blocks_iter = blocks.iter().copied().enumerate();
            let (_, last_block) = blocks_iter.next_back().unwrap_or_default();
            let Some((next_to_last_variant_index, next_to_last_block)) = blocks_iter.next_back()
            else {
                return Ok(StmtIf {
                    cond: true.to_expr(),
                    source_location,
                    blocks: [last_block.fold(state)?, Block::default()],
                });
            };
            let mut retval = StmtIf {
                cond: int_tag_expr
                    .cmp_eq(Expr::ty(int_tag_expr).from_int_wrapping(next_to_last_variant_index)),
                source_location,
                blocks: [next_to_last_block.fold(state)?, last_block.fold(state)?],
            };
            for (variant_index, block) in blocks_iter.rev() {
                retval = StmtIf {
                    cond: int_tag_expr
                        .cmp_eq(Expr::ty(int_tag_expr).from_int_wrapping(variant_index)),
                    source_location,
                    blocks: [
                        block.fold(state)?,
                        Block {
                            memories: Default::default(),
                            stmts: [Stmt::from(retval)][..].intern(),
                        },
                    ],
                };
            }
            Ok(retval)
        }
        match stmt {
            Stmt::Match(StmtMatch {
                expr,
                source_location,
                blocks,
            }) => match self.get_or_make_enum_type_state(Expr::ty(expr))? {
                EnumTypeState::TagEnumAndBody(_) => Ok(StmtMatch {
                    expr: Expr::<TagAndBody<Enum, DynSize>>::from_canonical(
                        Expr::canonical(expr).fold(self)?,
                    )
                    .tag,
                    source_location,
                    blocks: blocks.fold(self)?,
                }
                .into()),
                EnumTypeState::TagUIntAndBody(_) => {
                    let int_tag_expr = Expr::<TagAndBody<UInt, DynSize>>::from_canonical(
                        Expr::canonical(expr).fold(self)?,
                    )
                    .tag;
                    Ok(match_int_tag(self, int_tag_expr, source_location, blocks)?.into())
                }
                EnumTypeState::UInt(_) => {
                    let int_tag_expr =
                        Expr::<UInt>::from_canonical(Expr::canonical(expr).fold(self)?)
                            [..Expr::ty(expr).discriminant_bit_width()];
                    Ok(match_int_tag(self, int_tag_expr, source_location, blocks)?.into())
                }
                EnumTypeState::Unchanged => Ok(StmtMatch {
                    expr: expr.fold(self)?,
                    source_location,
                    blocks: blocks.fold(self)?,
                }
                .into()),
            },
            Stmt::Connect(_) | Stmt::If(_) | Stmt::Declaration(_) => stmt.default_fold(self),
        }
    }

    fn fold_stmt_match(&mut self, _v: StmtMatch) -> Result<StmtMatch, Self::Error> {
        unreachable!()
    }

    fn fold_canonical_type(
        &mut self,
        canonical_type: CanonicalType,
    ) -> Result<CanonicalType, Self::Error> {
        match canonical_type {
            CanonicalType::Enum(enum_type) => {
                Ok(match self.get_or_make_enum_type_state(enum_type)? {
                    EnumTypeState::TagEnumAndBody(ty) => ty.canonical(),
                    EnumTypeState::TagUIntAndBody(ty) => ty.canonical(),
                    EnumTypeState::UInt(ty) => ty.canonical(),
                    EnumTypeState::Unchanged => enum_type.canonical(),
                })
            }
            CanonicalType::Bundle(_)
            | CanonicalType::Array(_)
            | CanonicalType::UInt(_)
            | CanonicalType::SInt(_)
            | CanonicalType::Bool(_)
            | CanonicalType::Clock(_)
            | CanonicalType::AsyncReset(_)
            | CanonicalType::SyncReset(_)
            | CanonicalType::Reset(_) => canonical_type.default_fold(self),
        }
    }

    fn fold_enum_variant(&mut self, _v: EnumVariant) -> Result<EnumVariant, Self::Error> {
        unreachable!()
    }

    fn fold_enum_literal<T: EnumType>(
        &mut self,
        _v: ops::EnumLiteral<T>,
    ) -> Result<ops::EnumLiteral<T>, Self::Error>
    where
        T: Fold<Self>,
    {
        unreachable!()
    }

    fn fold_variant_access<VariantType: Type>(
        &mut self,
        _v: ops::VariantAccess<VariantType>,
    ) -> Result<ops::VariantAccess<VariantType>, Self::Error> {
        unreachable!()
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub enum SimplifyEnumsKind {
    SimplifyToEnumsWithNoBody,
    ReplaceWithBundleOfUInts,
    ReplaceWithUInt,
}

pub fn simplify_enums(
    module: Interned<Module<Bundle>>,
    kind: SimplifyEnumsKind,
) -> Result<Interned<Module<Bundle>>, SimplifyEnumsError> {
    module.fold(&mut State {
        enum_types: HashMap::new(),
        replacement_mem_ports: HashMap::new(),
        kind,
        name_id_gen: NameIdGen::default(),
    })
}
